//////////////////////////////////////////////
/// Classe snake : petit jeu JavaScript
///
/// Utilise : Objets, Event, Iterators, Array, DOM, Arrow functions
/// SeA - 2022
/////////////////////////////////////////////

class snake {
	constructor (x,y) {
		this.corps = [];	// Tableau des coordonnées de chaque morceau de snake
		this.biscuit = [x + 20, y + 30]; // Coordonnées d'origine du biscuit
		
		this.taille = 10; 	// Taille du carré de base (snake et biscuit)
		this.sens = 'H';	// Sens au démarrage du snake (vers le Haut)
		
		this.auto = 0;		// Timer d'avance automatique
		this.vitesse = 100;	// Temps (ms) entre 2 mouvements
		
		this.score = 0;		// Score réalisé
		
		this.ev = new Event("Score");	// Evènement généré à chaque nouveau score
		
		this.creerCorps(x,y);	// Initialisation du snake à l'écran
	}
	
	///////////////////////////////////////////////////////////
	/// Remplissage des coordonnées du snake au début du jeu(vertical)
	///////////////////////////////////////////////////////////
	creerCorps(x,y) {
		this.corps = [];
		for (let i=0; i<3; i++) {
			y = y + this.taille;
			this.corps.push([x, y]);
		}
		this.afficherCorps(); // Affichage
	}
	
	///////////////////////////////////////////////////////////
	/// Affichage à l'écran du snake et du biscuit
	///////////////////////////////////////////////////////////
	afficherCorps() {
		let snake = document.getElementById('snake');
		snake.innerHTML = "";
		this.corps.forEach( section => {
			snake.appendChild(this.ajouterElement(section[0], section[1], "red"));
		});
		
		// Affichage du biscuit
		snake.appendChild(this.ajouterElement(this.biscuit[0], this.biscuit[1], "green"));
	}
	
	////////////////////////////////////////////////////////////////
	/// Outil de création des <div> composant le snake ou le biscuit
	////////////////////////////////////////////////////////////////
	ajouterElement(x,y, couleur)
	{
		let uneDiv = document.createElement('div');
		uneDiv.style.position = "absolute";
		uneDiv.style.backgroundColor = couleur;
		uneDiv.style.height = this.taille + "px";
		uneDiv.style.width = this.taille + "px";
		uneDiv.style.left = x + "px";
		uneDiv.style.top = y + "px";
		return uneDiv;
	}
	
	////////////////////////////////////////////////////////////////
	/// En mode auto : le snake poursuit sa route
	////////////////////////////////////////////////////////////////
	avance(){ 
		let x = this.corps[0][0]; // Position de la tête
		let y = this.corps[0][1];
		
		if (this.sens == 'H') y -= this.taille;
		if (this.sens == 'B') y += this.taille;
		if (this.sens == 'D') x += this.taille;
		if (this.sens == 'G') x -= this.taille;
		
		this.corps.pop(); // Supprime la dernière position
		this.corps.unshift([x,y]); // Ajoute la tête à la nouvelle position
		
		this.controles();  // Vérification des obstacles + biscuit
		this.afficherCorps(); // Affichage nouvelle position
	}
	
	
	//////////////////////////////////////////////////////////////////
	/// Contrôle et alerte de dépassement de limite, de position biscuit, ...
	//////////////////////////////////////////////////////////////////
	controles() {
		// Dimensions du contener :
		let snake = document.getElementById('snake');
		let Xmax = snake.parentElement.offsetWidth ;
		let Ymax = snake.parentElement.offsetHeight - this.taille;
		
		// Test depassement des limites :
		let over = false;
		if ( this.corps[0][0] < 0 ) 	over = true	; // Gauche
		if ( this.corps[0][0] >= Xmax ) over = true; // Droite
		if ( this.corps[0][1] < 0 ) 	over = true; // Haut
		if ( this.corps[0][1] > Ymax )	over = true; // Bas
		if (over) this.reinit("Perdu !");
		
		// Test de retournement du snake contre lui même :
		this.corps.forEach( (section, i, snake) => {
			if ( i != 0 )
				if ( snake[0][0] == section[0] && snake[0][1] == section[1] ) 
					this.reinit("Le snake se mord la queue !!!");
		});
		
		// Test si on atteint un biscuit :
		if ( this.corps[0][0] == this.biscuit[0] &&  this.corps[0][1] == this.biscuit[1] ) {
			// Alors on rallonge le snake :
			this.modifSnake();
		}
	}

	///////////////////////////////////////////////////////////////
	/// Allongement du snake et augmentation vitesse
	///////////////////////////////////////////////////////////////	
	modifSnake() {	
		// Coordonnées de la queue:
		let x = this.corps[this.corps.length -1][0];
		let y = this.corps[this.corps.length -1][1];
		// Calcul des nouvelles coordonnées :
		if (this.sens == 'H') y -= this.taille;
		if (this.sens == 'B') y += this.taille;
		if (this.sens == 'D') x += this.taille;
		if (this.sens == 'G') x -= this.taille;
		// Ajoute à la queue :
		this.corps.push([x,y]); 
		
		// On met un autre biscuit ailleurs :
		this.changeBiscuit();
		
		// On modifie le score :
		this.score ++;
		// Génération de l'évènement "nouveau score" :
		document.dispatchEvent(this.ev); 
		// On accélère ...
		this.speedUp();
	}
		
	///////////////////////////////////////////////////////////////
	/// Calcul de la nouvelle position aléatoire du biscuit
	///////////////////////////////////////////////////////////////
	changeBiscuit() {
		let snake = document.getElementById('snake');
		let Xmax = snake.parentElement.offsetWidth ;
		let Ymax = snake.parentElement.offsetHeight - this.taille;
		
		// Calcul de l'interval de position pour coïncider avec le déplacement snake
		let nbX = Math.floor(Xmax / this.taille);
		let nbY = Math.floor(Ymax / this.taille);
		// Nouvelle position alléatoire
		let x = Math.floor(Math.random() * nbX) * this.taille;
		let y = Math.floor(Math.random() * nbY) * this.taille;
		
		this.biscuit[0] = x;
		this.biscuit[1] = y;
	}
	
	///////////////////////////////////////////////////////////////
	/// Accélération du mouvement en reprogrammant le timer
	///////////////////////////////////////////////////////////////
	speedUp() {
		if ( this.vitesse > 10) this.vitesse -= 5;
		if (this.auto)	{
			clearInterval(this.auto);
			this.auto = setInterval(() => this.avance(), this.vitesse);
		}
	}
	
	///////////////////////////////////////////////////////////////
	/// Réinitialisation du jeu si la partie est perdue
	///////////////////////////////////////////////////////////////
	reinit (msg) {
		this.startStop();
		alert(msg);
		this.creerCorps(100,10);
		this.score = 0;
		this.vitesse = 100;
	}
	
	///////////////////////////////////////////////////////////////
	/// Fonction permettant d'attendre l'évènement "nouveau score"
	/// Méthode publique : Appel par le programme principal
	///////////////////////////////////////////////////////////////
	onScore(callback) { // addEventListener interne
		document.addEventListener("Score", callback);
		window.focus();
	}
	
	///////////////////////////////////////////////////////////////
	/// Donner une nouvelle direction au snake
	/// Méthode publique : Appel par le programme principal
	///////////////////////////////////////////////////////////////
	tourne(sens) {
		this.sens = sens;
	}
	
	///////////////////////////////////////////////////////////////
	/// Démarrage du mouvement automatique du snake
	/// Méthode publique : Appel par le programme principal
	///////////////////////////////////////////////////////////////
	start() {
		this.auto =  setInterval(()=>this.avance(), this.vitesse);
	}
	
	///////////////////////////////////////////////////////////////
	/// Pause / reprise du mouvement du snake
	/// Méthode publique : Appel par le programme principal
	///////////////////////////////////////////////////////////////
	startStop() {
		if (this.auto)	{
			clearInterval(this.auto); this.auto = 0;
		}
		else 
			this.auto = setInterval(() => this.avance(), this.vitesse);
	}
	
	///////////////////////////////////////////////////////////////
	/// Obtenir le nouveau score
	/// Méthode publique : Appel par le programme principal
	///////////////////////////////////////////////////////////////
	getScore() {
		return this.score;
	}
} // Fin de la classe snake